#define MAIN_C TRUE
/* ========================================================================
 * Name - main.c
 *
 * Description - mains for woman, wwhatis, wapropos
 *   Woman -Wonderful Online MANual.
 *   Main module, containing main program, searching of
 *   directories, nroffing, paging, etc.
 *
 * $Id: main.c,v 1.3 1992/09/16 22:33:37 penneyj Exp $
 *
 * $Log: main.c,v $
 * Revision 1.3  1992/09/16  22:33:37  penneyj
 * misc cleanups
 *
 * Revision 1.2  1992/09/15  22:37:20  penneyj
 * misc facelift, preparing for distribution on alt.sources
 *
 * Revision 1.1  1992/09/15  22:18:47  penneyj
 * Initial revision
 *
 *========================================================================
 */

#include "general.h"
#include "config.h"

#include "match.h"
#include "list.h"
#include "utils.h"
#include "dowh.h"
#include "format.h"

#include "main.h" /* this file's interface */

static ListType secpart[NUMSECTS];

#if defined(FLG_PROTOTYPE)
void warning(const char *s0, const void *s1)
#else
void warning(s0, s1)
const char *s0;
const void *s1;
#endif
{
  fprintf(stderr, "%s: ", whoami);
  fprintf(stderr, s0 /* pattern */, s1 /* value */);
  fprintf(stderr, "\n");
}


/*
 *  Removes tmpfiles.
 */

#if defined(FLG_PROTOTYPE)
void cleanup(void)
#else
void cleanup() 
#endif
{
  ListType item;
  char *this;

  LIST_DO (tmpfiles) {
    this = (*item)->key;
    if (unlink(this)) {
      warning("Couldn't remove %s", this);
      perror(this);
      }
    }
}

/*
 *  To catch aborts from the user.
 */
#if defined(FLG_PROTOTYPE)
void caught(int sig)
#else
void caught(sig)
int sig;
#endif
{
  fprintf(stderr, "  aborted (Sorry)\n");
  if (childtokill != 0) {
    kill(childtokill, SIGTERM);
    waitamoment();
    DOWAIT;
    kill(childtokill, SIGKILL);
    waitamoment();
    DOWAIT;
    }
  cleanup();
  unlink(notyet); /* CCL */
  die(ABORTED);
}


/* 
 * insect() and insect2() determines if the section is right 
 */
#if defined(FLG_PROTOTYPE)
BoolType insect(const char *str)
#else
BoolType insect(str)
const char *str;
#endif
{
  ListType item;

  LIST_DO (sections) {
    if (strncmp((*item)->key, str, strlen((*item)->key)) == 0) 
      return TRUE;
    }
  return FALSE;
}


#if defined(FLG_PROTOTYPE)
BoolType insect2(char a)
#else
BoolType insect2(a)
char a;
#endif
{
  ListType item;

  if (a == '\0') 
    return FALSE;

  LIST_DO (sections) {
    if ((*item)->key[0] == a) 
      return TRUE;
    }
  return FALSE;
}


/*
 *  Is the file one of those looked for?
 */

#if defined(FLG_PROTOTYPE)
static int lookforpage(const char *filename, const char *where)
#else
static int lookforpage(filename, where)
     const char *filename, *where;
#endif
{
  ListType item;
  char *sect, *this;
  static const char *lastout = NULL;
  
  sect = strrchr(filename, '.');
  if (sect != NULL)
    sect++;
  else {
    if (debug_f && (where[0] != '.')) {
      if (where != lastout)
        fprintf(stderr, "In directory %s:\n", where);
      lastout = where;
      warning( "No dot in filename %s" , filename);
      }
    return (int)NO_PAGE;
    }

  /* System V packed cat-file */
  if (strcmp(sect, "z") == 0) {
    LIST_DO (keywords) {
      this = (*item)->key;
      if (matchKey(filename, this)) 
	return (int)PACKED_CAT_PAGE;
      }
    return (int)NO_PAGE;
    }

  /* Pseudo-System V compressed cat-file */
  if (strcmp(sect, "Z") == 0) {
    LIST_DO (keywords) {
      this = (*item)->key;
      if (matchKey(filename, this)) 
	return COMPRESSED_CAT_PAGE;
      }
    return (int)NO_PAGE;
    }
  
  if (strchr(SECTALLS, *sect) == NULL) 
    return (int)NO_PAGE;
  
  if (*sections != NULL) 
    if (!insect(sect)) 
      return (int)NO_PAGE;
  LIST_DO (keywords) {
    this = (*item)->key;
    if (matchKey(filename, this)) {
      char *junk;

      /* check for hpux style: man1.Z/sh.1, compressed nroff file */
      junk = strchr(where, '.');
      if (junk == NULL)
	goto NOT_COMPRESSED;
      if (strcmp(&junk[1], "Z") != 0)
	goto NOT_COMPRESSED;
      return (int)ZCAT_NROFF_PAGE;

NOT_COMPRESSED:
      /* to do: what if cat page but no man page? */
      return (int)NROFF_PAGE;
      }
    }
  return (int)NO_PAGE;
}

#if defined(FLG_PROTOTYPE)
static BoolType chstatf(struct stat *buf)
#else
static BoolType chstatf(buf)
     struct stat *buf;
#endif
{
  return (int)S_ISREG(buf->st_mode);
}

#if defined(FLG_PROTOTYPE)
static BoolType insfile(const char *fullname, const char *shortname,
    int nametype, void *divinfo)
#else
static BoolType insfile(fullname, shortname, nametype, divinfo)
     const char *fullname, *shortname;
     int nametype;
     void *divinfo;
#endif
{
  char *manfnam, *catfnam, *catdir, *treetop;

  treetop = ((char **)divinfo)[0];     
  catdir = ((char **)divinfo)[1];     

  if (verbos_f) {
    fprintf(stderr, "Found: %s (type %d)\n", fullname, nametype);
    if (catdir != NULL)
      fprintf(stderr, "       catdir %s\n", catdir);
    }

  switch ((manPageEType)nametype) {
    case NROFF_PAGE:
    case PACKED_CAT_PAGE:
    case COMPRESSED_CAT_PAGE:
    case ZCAT_NROFF_PAGE:
      catfnam = NULL;
      if (catdir != NULL) {
        char *shortName = strrchr(fullname, '/');

	if (shortName != NULL)
	  catfnam = makeCatName(fullname, catdir, &shortName[1]);
	}
      manfnam = (char *)MyMalloc(strlen(fullname) + EXTRASPACE);
      strcpy(manfnam, fullname);
      add3list(matchingFiles, manfnam, catfnam, treetop,
          (manPageEType)nametype);
      return TRUE;
      break;

    default:
      fprintf(stderr, "Unknown nametype for name %s\n", fullname);
      break;
    }

  return FALSE;
}

#if defined(FLG_PROTOTYPE)
static void initialize(void) 
#else
static void initialize() 
#endif
{
  umask(0);    /* This program should only create world-writable files ! */
  notyet = NULL;
  childtokill = 0;
  keywords = newlist();
  matchingFiles = newlist();
  showfiles = newlist();
  searchdirs = newlist();
  sections = newlist();
  searchtrees = newlist();
  tmpfiles = newlist();
}

/*
 *  parseargs() does the right thing.
 */
#if defined(FLG_PROTOTYPE)
static void parseargs(int argc, const char **argv)
#else
static void parseargs(argc, argv)
int  argc;
const char **argv;
#endif
{
  BoolType errorflag;
  int  i;
  char *extrarg;
  const char **nargv;
  int nargc;
  
  opterr = 0;
  reform_f = nopag_f = debug_f = verbos_f = every_f = only1_f = FALSE;
  manmacros = NULL;
  errorflag = FALSE;

  whoami = argv[0]; /* Take a copy. */
  if (strchr(whoami, '/') != NULL)
    whoami = 1 + strrchr(whoami, '/'); /* skip explicit path */

#if 0
  if ((STREQUAL(whoami, "man")) || (STREQUAL(whoami, MAN_NAME)) || 
   (STREQUAL(whoami, "human")))
#endif
  if (STREQUAL(whoami, MAN_NAME))
    whattodo = INVOKED_MAN;
  else
  if (STREQUAL(whoami, APROPOS_NAME))
    whattodo = INVOKED_APROPOS;
  else
  if (STREQUAL(whoami, WHATIS_NAME))
    whattodo = INVOKED_WHATIS;
  else
    whattodo = INVOKED_MAN;
#if 0
  if (STREQUAL(whoami, "a.out")) {
    fprintf(stderr, "This is an outrage! You can't call me that!\n");
    die(USRERROR); /* Verree funnee */
    }
  else {
    fprintf(stderr, "I'm having identity problems. %s %s.\n",
        "Please call me something else than", whoami);
    die(USRERROR); /* Also verree funnee */
    }
#endif

  if (argc < 2) 
    prusage();               /*  Might do 'man man'              */

  extrarg = getenv("MANOPT");
  if (extrarg != NULL) {
    int j;

    nargv = MyMalloc((strlen(extrarg) + argc) * sizeof(char *));
    nargv[0] = "";
    nargc = 1;
    do {
      nargv[nargc ++] = extrarg;
      extrarg = strchr(extrarg, ' ');
      if (extrarg != NULL) {
	*extrarg= '\0';
	extrarg ++;
	if (*extrarg != '-' && *extrarg != '\0') {
	  warning("Illegal argument in MANOPT: %s", extrarg);
	  break;
	  }
        }
      } while (extrarg != NULL);
    for (j = 1; j < argc; j++)
      nargv[nargc ++] = argv[j];
    }
  else {
    nargv = argv;
    nargc = argc;
    }
  
  while ((i = getopt(nargc,
#if defined(FLG_SEQUENT_UNIX)
	(char **)
#endif
	nargv,
	"ufkdvaiX1trnT:")) != EOF)
    switch (i) {
      case 'u':
        prusage();
	/* NOTREACHED */

      case 'f':        /* 'man -f' is an undocumented way of */
                       /* calling whatis(1) */
        if (whattodo == INVOKED_APROPOS) 
	  prusage();
        whattodo = INVOKED_WHATIS;
        break;

      case 'k':        /* 'man -k' is an undocumented way of */
                       /* calling apropos(1) */
        if (whattodo == INVOKED_WHATIS) 
	  prusage();
        whattodo = INVOKED_APROPOS;
        break;

      case 'd':
        debug_f = TRUE;
        break;
      case 'v':
        verbos_f = TRUE;
        break;
      case 'a':
        every_f = TRUE;
        break;
      case 't':
        nopag_f = TRUE;
        break;
      case 'r':
        reform_f = TRUE;
        break;
      case 'i':
        igncas_f = TRUE;
        kilcap_f = FALSE;
        break;
      case 'X':
        kilcap_f = TRUE;
        igncas_f = FALSE;
        break;
      case '1':
        only1_f = TRUE;
        break;
      case 'n': /* turn off all strange options */
        reform_f = nopag_f = only1_f = every_f = debug_f = verbos_f = FALSE;
        break;
      case 'T':
        manmacros = optarg;
        if (access(manmacros, R_OK) != 0) {
          warning("manmacros %s is not readable", manmacros);
          manmacros = NULL;
          }
        break;

      case '?':
      default:
        errorflag = TRUE;
      }
  if (errorflag || optind >= nargc) 
    prusage();
  if (whattodo == INVOKED_MAN) {
    /* Parsing section options only in 'man' mode */
    if (strchr(SECTNUMS, nargv[optind][0]) != NULL) {
      ADDTOLIST(sections, nargv[optind]);
      if (verbos_f)
        warning("Searching only section %s", nargv[optind]);
        optind++;
      }
    else
    if (STREQUAL(nargv[optind], "local"))
      ADDTOLIST(sections, "l");
    else
    if (STREQUAL(nargv[optind], "new"))
      ADDTOLIST(sections, "n");
    else
    if (STREQUAL(nargv[optind], "public"))
      ADDTOLIST(sections, "p");
    else
    if (STREQUAL(nargv[optind], "old"))
      ADDTOLIST(sections, "o");
    }
  for ( ; optind < nargc; optind++) {
    if (optind == 0) /* don't add program name as keyword! */
      continue;
    if (strchr(nargv[optind],'/') != NULL) {
      /* ADDTOLIST(matchingFiles, nargv[optind]); */
      add3list(matchingFiles, nargv[optind], NULL, NULL, NROFF_PAGE);
      }
    else
      ADDTOLIST(keywords, nargv[optind]);
    }
  if (manmacros == NULL)
    manmacros = MANMACROS;
  if (*keywords == NULL && *matchingFiles == NULL)
    prusage();
}


#if defined(FLG_PROTOTYPE)
static void parsemanpath(void)
#else
static void parsemanpath() 
#endif
{
  char *manpath;
  char *path, *extraspace;

  manpath = getenv("MANPATH");
  if (manpath == NULL)
    manpath = (char *)MANPATH;
  extraspace = MyMalloc(strlen(manpath) + EXTRASPACE);
  strcpy(extraspace, manpath);
  manpath = extraspace;

  while (manpath != NULL && (*(path = manpath)) != '\0') {
    struct stat statbuf;

    manpath = strchr(path, ':');
    if (manpath != NULL)
      manpath[0] = '\0';
    /* only add if it's really there */
    if (stat(path, &statbuf) == 0 && S_ISDIR(statbuf.st_mode))
	ADDTOLIST(searchtrees, path);
    else
    if (verbos_f)
        warning("Figured this was not a directory: %s", path);
      
    if (manpath != NULL)
      manpath++;
    }
}


/*
 *  Is this a directory we want to search ? 
 *  Retruns the number of the section(1-12) for true.
 *  Returns 0 if it's something else.
 */
#if defined(FLG_PROTOTYPE)
static int okdir(const char *name, const char *unused)
#else
static int okdir(name, unused)
const char *name;
const char *unused;
#endif
{
  char *temp;
  const char *thesections = SECTALLS;

  if ( name[3] == '\0' || strncmp(name, "man", 3) != 0 ) 
    return 0;
  if (*sections != NULL)
    if (!insect2(name[3])) 
      return 0;
  temp = strchr(thesections, name[3]);
  if (temp != NULL) 
    return temp - thesections + 1;
  return 0;
}


#if defined(FLG_PROTOTYPE)
static BoolType chstatd(struct stat *buf)
#else
static BoolType chstatd(buf)
     struct stat *buf;
#endif
{
  return (BoolType)S_ISDIR(buf->st_mode);
}

#if defined(FLG_PROTOTYPE)
static int insdir(const char *fullname, const char *shortname,
    int nametype, void *divinfo)
#else
static int insdir(fullname, shortname, nametype, divinfo)
     const char *fullname, *shortname;
     int nametype;
     void *divinfo;
#endif
{
  char *mandirnam, *catdirnam, *treetop;
  struct stat statbuf;

  treetop = ((char **)divinfo)[0];
  nametype --;
  
  if (nametype == OLD && !every_f && !insect2('o')) {
    if (debug_f)
      warning("Thrown away section %s", shortname);
    return (int)FALSE;
    }

  mandirnam = (char *)MyMalloc(strlen(fullname) + EXTRASPACE);
  strcpy(mandirnam, fullname);

  catdirnam = (char *)MyMalloc(strlen(fullname) + EXTRASPACE);
  sprintf(catdirnam, "%s%s%s%s", treetop, DIRSEP, "cat", &shortname[3]);
  
  if (stat(catdirnam, &statbuf) != 0)
    catdirnam = NULL;
  else
  if (!S_ISDIR(statbuf.st_mode))
    catdirnam = NULL;

  push3list(secpart[nametype], mandirnam, catdirnam, treetop);
  return (int)TRUE;
}


/* For each tree in manpath, find interesting directories. */
#if defined(FLG_PROTOTYPE)
static void treestodirs(void)
#else
static void treestodirs() 
#endif
{
  ListType item;
  int  counter;
  char *this;
  char **divinfo;
  int totalfound;
  
  divinfo = (char **)MyMalloc(1*sizeof(char **));
  
  for (counter = 0; counter < NUMSECTS; counter++)
    secpart[counter] = addnull(searchdirs);

  LIST_DO (searchtrees) {
    this = (*item)->key;
    divinfo[0] = this;
    
    totalfound = dirsearch(this, okdir, chstatd, insdir, (void *)divinfo);

    if (totalfound == 0) {
      add3list(searchdirs, this, NULL, this, NO_PAGE);
      if (verbos_f && (this[0] != '.'))
        warning("Figured this was not a tree: %s", this);
      }
    }	  
  cleannulls(searchdirs);     
  FREE(divinfo);     
}

#if defined(FLG_PROTOTYPE)
static void searchemall(void)
#else
static void searchemall() 
#endif
{
  ListType item;
  char **divinfo;

  divinfo = (char **)MyMalloc(2 * sizeof(char **));
  LIST_DO (searchdirs) {
    divinfo[1] = (*item)->catname;
    divinfo[0] = (*item)->dirname;
    dirsearch((*item)->key, lookforpage, chstatf, insfile, (void *)divinfo);
    }	  
  FREE(divinfo);     
}


/* Interface to the pager. */
#if defined(FLG_PROTOTYPE)
static void showthem(void)
#else
static void showthem() 
#endif
{
  const char **args;
  ListType item;
  char *this;
  char *pager;
  int  counter;

  counter = 0;
  if (*showfiles != NULL) {
    if (!isatty(fileno(stdout))) /* just pipe them to stdout */
	this = (char *)"/bin/cat";
    else {
      this = getenv("PAGER");
	if (this == NULL)
	  this = (char *)PAGER;
      }

    pager = MyMalloc(strlen(this) + EXTRASPACE);
    strcpy(pager, this);

    args = MyMalloc(strlen(pager) + EXTRASPACE + noofitems(showfiles));
    args[counter ++] = "Man Pager";

    this = pager;
    while ((this = strchr(this, ' ')) != NULL) {
      *this = '\0';
      args[counter ++] = ++ this;
    }

    LIST_DO (showfiles) {
      this = (*item)->key;
      args[counter ++] = this;
      }
    args[counter] = NULL;
    childtokill = fork();
    if (childtokill == -1) {
      perror("fork of pager failed");
      die(OUTSIDERROR);
      }
    else
    if (childtokill == 0) {
      if (nopag_f) {
        /* Better printlist needed here. */
        printlist(showfiles);
	die(OKEXIT);
	}
      execvp(pager,
#if defined(FLG_HPUX_UNIX)
	(char **)
#endif
	args);
      perror("Execvp of pager failed");
      die(OUTSIDERROR);
      }
    wait(NULL);
    childtokill = 0;
    return;
    }

  /* This point should no longer be reached! */
  fprintf(stderr, "! No manual entry for");
  LIST_DO (keywords) {
    fprintf(stderr, " %s", (**item).key);
    }
  fprintf(stderr, ".\n");
}

#if defined(FLG_PROTOTYPE)
int main(int argc, const char **argv)
#else
int main(argc, argv)
int  argc;
const char **argv;
#endif
{
  signal(SIGINT, caught);
  initialize();
  parseargs(argc, argv);
  parsemanpath();
  switch (whattodo) {
    case INVOKED_MAN:
      if (*keywords != NULL) {
        treestodirs();
	searchemall();
	if (*matchingFiles == NULL) {
	  fprintf(stderr,"%s: Didn't find anything.\n", whoami);
	  die(OKEXIT);
	  }
	}
      formatfiles();
      break;

    case INVOKED_WHATIS:
    case INVOKED_APROPOS:
      dowhatis();
      break;
    }
  showthem();
  cleanup();
  return 0;
}
